--
-- DustWheels
-- Specialization for DustWheels
--
-- This script adds to every wheel:
--		- a dust particle system 
--  	- an area for destruction of fruit
--		- dirt track, when leaving a field  (?)
--	
-- @author  	fruktor (www.eifok-team.de)
-- @date:		26/12/12
-- @version:	v0.1
-- @history:	v0.1 - initial implementation
--
-- free for noncommerical-usage
--
	
local ret = FruitUtil.registerFruitType("dust", false, false, false, 1, 1, 1, 1, 1, nil);
print("  DustWheels :: register fruittype dust -> for wheel-PS besides fields      ret = "..tostring(ret));


DustWheels = {};

function DustWheels.prerequisitesPresent(specializations)
    return true; --SpecializationUtil.hasSpecialization(Motorized, specializations);
end;

function DustWheels:load(xmlFile)
	--print(" ============ function DustWheels:load(xmlFile) ============ ");
	

	self.cutFruitArea_forced = DustWheels.cutFruitArea_forced;
	
	--[[
	print("----------------------------------------------------------");
	for index,fruit in pairs(g_currentMission.fruits) do
		print("________ " .. tostring(index) .. " " .. tostring(fruit) );
		local fruitDesc = FruitUtil.fruitIndexToDesc[index];
		print("fruitDesc="..tostring(fruitDesc).."   "..tostring(fruitDesc.needsSeeding));
		for i,j in pairs(fruitDesc) do
			print(tostring(i).." "..tostring(j));
		end;
	end;	
	]]--

	if SpecializationUtil.hasSpecialization(Steerable, self.specializations) then
		self.dsw_directory = Steerable.dsw_directory;
	else
		self.dsw_directory = Attachable.dsw_directory;
	end

	self.isCrawler = SpecializationUtil.hasSpecialization(Crawler, self.specializations);
	
	self.hasWheels = false;
	if self.wheels ~= nil then
		if self.wheels[1] ~= nil and self.wheels[2] ~= nil then
			self.hasWheels = true;
		end
	end
	
	if self.isCrawler then
		self.hasWheels = false;
		print("self.isCrawler <----------------------");
		--for i,j in pairs(self) do
		--	print(tostring(i).." "..tostring(j));		
		--end;
		--[[print("self.crawlers >>>>>>>>>>>>>>>>>>>>> ");
		for i,j in pairs(self.crawlers) do
			print(tostring(i).." "..tostring(j));
			for k,l in pairs(j) do
				print(":_:"..tostring(k).." "..tostring(l));
			end
			for k,l in pairs(j.rotatingParts) do
				print(":r:"..tostring(k).." "..tostring(l));
				for v,w in pairs(l) do
					print(":r.:"..tostring(v).." "..tostring(w));
				end
				local parentNode = getParent(l.node);
				print("-->   getNumOfChildren(parentNode)="..tostring(getNumOfChildren(parentNode)));
				for run=0,getNumOfChildren(parentNode)-1 do
					childId = getChildAt(parentNode, run);
					print("-->   childId="..tostring(childId));
					if childId == l.node then
						print("HIT!");
						break;
					end;
				end
				
			end
		end;	
		]]--
		return;
	end;

	self.rcCallBack = DustWheels.rcCallBack;
	
	self.dtTmp = 0;
	self.dtDelta = 400;	
	self.drivingParticleThreshold = 0.05;

	self.destructAreas = {};
	self.destructAreasOld = {};
	self.destructThresh = 0.5;	
	
	--# tracks
	self.wt = {};
	self.wt.fileDry = Utils.convertToNetworkFilename( self.dsw_directory .. "dirtDry.i3d" ); --Utils.getFilename("dirtDry.i3d", self.baseDirectory) ); 
	self.wt.tracks = {};
	
	--#
	local hasManualDefinition = hasXMLProperty( xmlFile, string.format("vehicle.dustwheels.dustwheel(0)#size") );
	--print("hasManualDefinition="..tostring(hasManualDefinition));

	
	local dir = getUserProfileAppPath() .. "savegame"..g_currentMission.missionInfo.savegameIndex;
	createFolder(dir);	
	self.tmpXmlFile = dir .. "/tmpXmlFile.xml";

	createXMLFile("tmpXmlFile", self.tmpXmlFile, "tmpXmlFile");		
	local xmlFileT = {};
	xmlFileT = io.open (self.tmpXmlFile, "w");	
	xmlFileT:write('<?xml version="1.0" encoding="utf-8" standalone="no" ?>\n<tmpConfig>\n    <drivingParticles>\n');

	local i3dname = getXMLString(xmlFile, "vehicle.filename");
	local filename2 = Utils.getFilename(i3dname, self.baseDirectory);	
	local i3dID = {};
	if not hasManualDefinition then
		i3dID = loadXMLFile( "i3dfile", filename2 );
	end;
	
	local tmpDustnodes = {};
	local tmpDustNodeRefString = {};
	
	
	--print("-------------------->");
	for i,w in pairs(self.wheels) do
		--print("------>> Wheel "..i);
		local maxDim = 0;
		local centerOffset = 0;
		
		if hasManualDefinition then
			
			local size = getXMLFloat(xmlFile, string.format("vehicle.dustWheels.dustWheel(%d)#size", i-1) );
			if size ~= nil then
				maxDim = size;
			end
						
		else

			local str = getXMLString(xmlFile, string.format("vehicle.wheels.wheel(%d)#repr", i-1) );
			local str2 = getXMLString(xmlFile, string.format("vehicle.wheels.wheel(%d)#driveNode", i-1) );
			if str2 ~= nil then
				str = str2;
			end		
			
			local id1 = 0;
			local id2 = 0;
			local id3 = -1;
			local begin = string.find(str, ">", nil, true);
			if begin ~= nil then

				if begin == 2 then
					id1 = string.byte(str,1)-48;
				else
					local f = 1;
					for r=begin-1,1,-1 do
						id1 = id1 + (string.byte(str,r)-48)*f;
						f = f*10;
					end
				end
			else
				begin = 0;
			end
			
			
				local s = string.find(str, "|", nil, true);
				
				if s then
				
					if s-begin == 2 then
						id2 = string.byte(str, begin+1)-48;
					else
						local f = 1;
						for r=s-1,begin+1,-1 do
							id2 = id2 + (string.byte(str,r)-48)*f;
							f = f * 10;
						end
					end

					if string.len(str) == s + 1 then
						id3 = string.byte(str, string.len(str))-48;
					else
						local f = 1;
						for r=string.len(str),s+1,-1 do
							id3 = id3 + (string.byte(str,r)-48)*f;
							f = f * 10;
						end
					end;
					
				else
				
					if string.len(str)-begin == 2 then
						id2 = string.byte(str, begin+1)-48;
					else
						local f = 1;
						for r=string.len(str),begin+1,-1 do
							id2 = id2 + (string.byte(str,r)-48)*f;
							f = f * 10;
						end
					end		
					
				end			

			--print("FROM string: "..str.." TO:" .. id1 .." "..id2.." "..id3);
			
			local name;
			local shapeId;
			local shape = false;
			
			-- lvl 1
			local nodeIdList = {};
			local objList = {};
			local run = 0;
			while true do
				if hasXMLProperty( i3dID, string.format("i3D.Scene.Shape(%d).Shape(%d)#nodeId", id1, run) ) then
					local e={};
					e.idx = run;
					local nodeId = getXMLInt( i3dID, string.format("i3D.Scene.Shape(%d).Shape(%d)#nodeId", id1, run) );
					e.nodeId = nodeId;
					e.typ = 'Shape';
					table.insert( nodeIdList, nodeId );
					table.insert( objList, e );
					run = run + 1;
				else
					break;
				end;
			end
			run = 0;
			while true do
				if hasXMLProperty( i3dID, string.format("i3D.Scene.Shape(%d).TransformGroup(%d)#nodeId", id1, run) ) then
					local e={};
					e.idx = run;
					local nodeId = getXMLInt( i3dID, string.format("i3D.Scene.Shape(%d).TransformGroup(%d)#nodeId", id1, run) );
					e.nodeId = nodeId;
					e.typ = 'TransformGroup';
					table.insert( nodeIdList, nodeId );
					table.insert( objList, e );
					run = run + 1;
				else
					break;
				end;
			end
			run = 0;
			while true do
				if hasXMLProperty( i3dID, string.format("i3D.Scene.Shape(%d).Camera(%d)#nodeId", id1, run) ) then
					local e={};
					e.idx = run;
					local nodeId = getXMLInt( i3dID, string.format("i3D.Scene.Shape(%d).Camera(%d)#nodeId", id1, run) );
					e.nodeId = nodeId;
					e.typ = 'Camera';
					table.insert( nodeIdList, nodeId );
					table.insert( objList, e );
					run = run + 1;
				else
					break;
				end;
			end		
			--print("#################### 1");
			--for idx,obj in pairs(objList) do
			--	print(tostring(idx).." "..tostring(obj.idx).." "..tostring(obj.nodeId).." "..tostring(obj.typ));
			--end;
			--print("####################");
			
			--str = '';
			--for idx,id in pairs(nodeIdList) do
			--	str = str .. " ".. tostring(id);
			--end;
			--print(str);			
			table.sort(nodeIdList)
			--str = '';
			--for idx,id in pairs(nodeIdList) do
			--	str = str .. " ".. tostring(id);
			--end;
			--print(str);	
			
			local index2 = -1;
			for idx,obj in pairs(objList) do
				if obj.nodeId == nodeIdList[id2+1] then
					index2 = idx;
					break;
				end;
			end;
			--print(index2);
			--print(nodeIdList[index2]);
			
			local index21 = objList[index2 ].idx;
			--print(index21);
				
			--print("########FIN#########");
			
			if objList[index2 ].typ == 'TransformGroup' then
			--	print("-----> TG as wheel "..i);
				shape = false;
			else
			--	print("-----> Shape as wheel "..i);
				shape = true;
			end;
			
			if shape then
				name = getXMLString( i3dID, string.format("i3D.Scene.Shape(%d).Shape(%d)#name", id1, objList[index2 ].idx )  );
				shapeId = getXMLString( i3dID, string.format("i3D.Scene.Shape(%d).Shape(%d)#shapeId", id1, objList[index2 ].idx )  );
			end
			
			-- lvl 2
			if id3 ~= -1 then
			
				local nodeIdList2 = {};
				local objList2 = {};
				local run = 0;
				while true do
					if hasXMLProperty( i3dID, string.format("i3D.Scene.Shape(%d).%s(%d).Shape(%d)#nodeId", id1, objList[index2 ].typ, objList[index2 ].idx, run) ) then
						local e={};
						e.idx = run;
						local nodeId = getXMLInt( i3dID, string.format("i3D.Scene.Shape(%d).%s(%d).Shape(%d)#nodeId", id1, objList[index2 ].typ, objList[index2 ].idx, run) );
						e.nodeId = nodeId;
						e.typ = 'Shape';
						table.insert( nodeIdList2, nodeId );
						table.insert( objList2, e );
						run = run + 1;
					else
						break;
					end;
				end
				run = 0;
				while true do
					if hasXMLProperty( i3dID, string.format("i3D.Scene.Shape(%d).%s(%d).TransformGroup(%d)#nodeId", id1, objList[index2 ].typ, objList[index2 ].idx, run) ) then
						local e={};
						e.idx = run;
						local nodeId = getXMLInt( i3dID, string.format("i3D.Scene.Shape(%d).%s(%d).TransformGroup(%d)#nodeId", id1, objList[index2 ].typ, objList[index2 ].idx, run) );
						e.nodeId = nodeId;
						e.typ = 'TransformGroup';
						table.insert( nodeIdList2, nodeId );
						table.insert( objList2, e );
						run = run + 1;
					else
						break;
					end;
				end
				run = 0;
				while true do
					if hasXMLProperty( i3dID, string.format("i3D.Scene.Shape(%d).%s(%d).Camera(%d)#nodeId", id1, objList[index2 ].typ, objList[index2 ].idx, run) ) then
						local e={};
						e.idx = run;
						local nodeId = getXMLInt( i3dID, string.format("i3D.Scene.Shape(%d).%s(%d).Camera(%d)#nodeId", id1, objList[index2 ].typ, objList[index2 ].idx, run) );
						e.nodeId = nodeId;
						e.typ = 'Camera';
						table.insert( nodeIdList, nodeId );
						table.insert( objList, e );
						run = run + 1;
					else
						break;
					end;
				end				
				--print("#################### 2 ");
				--for idx,obj in pairs(objList2) do
				--	print(tostring(idx).." "..tostring(obj.idx).." "..tostring(obj.nodeId).." "..tostring(obj.typ));
				--end;
				
				--str = '';
				--for idx,id in pairs(nodeIdList2) do
				--	str = str .. " ".. tostring(id);
				--end;
				--print(str);		
				table.sort(nodeIdList2)
				--str = '';
				--for idx,id in pairs(nodeIdList2) do
				--	str = str .. " ".. tostring(id);
				--end;
				--print(str);	
				
				local index3 = -1;
				for idx,obj in pairs(objList2) do
					if obj.nodeId == nodeIdList2[id3+1] then
						index3 = idx;
						break;
					end;
				end;
				--print(index3);
				--print(nodeIdList2[index3]);
				
				local index31 = objList2[index3 ].idx;
				--print(index31);

				
				local index31 = -1;
				for idx,obj in pairs(objList2) do
					if obj.nodeId == nodeIdList2[index3] then
						index31 = idx;
						break;
					end;
				end;				
				--print(index31);
				--print(objList2[index3 ].idx);
				
				--print("########FIN#########");
				
				if objList2[index3 ].typ == 'TransformGroup' then
					--print("-----> TG as wheel "..i);
					shape = false;
				else
					--print("-----> Shape as wheel "..i);
					shape = true;
				end;
				
				if shape then
					name = getXMLString( i3dID, string.format("i3D.Scene.Shape(%d).%s(%d).Shape(%d)#name", id1, 
						objList[index2 ].typ, objList[index2 ].idx,
						objList2[index3 ].idx ) );
					shapeId = getXMLString( i3dID, string.format("i3D.Scene.Shape(%d).%s(%d).Shape(%d)#shapeId", id1, 
						objList[index2 ].typ, objList[index2 ].idx,
						objList2[index3 ].idx ) );
				end
			
			end;
			--print("name="..tostring(name).." shapeId="..tostring(shapeId));	
			
			if not shape then
				maxDim = -1;			
			end;		
			
			if maxDim ~= -1 then 
				local run = 0;
				while true do
					local tmpStr = getXMLString( i3dID, string.format("i3D.Shapes.IndexedTriangleSet(%d)#shapeId", run));
					if tmpStr == nil then
						--print("STOP at "..run);
						break;
					end;
					if string.len(shapeId) == string.len(tmpStr) then
						if string.find(shapeId, tmpStr, nil, true) then
							--print("HIT -> "..run);
							local vertCnt = getXMLInt( i3dID, string.format("i3D.Shapes.IndexedTriangleSet(%d).Vertices#count", run));
							--print("vertCnt="..tostring(vertCnt));
							local maxX = 0;
							local minX = 100;
							for v=1,vertCnt do
								local x, y, z = Utils.getVectorFromString(getXMLString(i3dID, string.format("i3D.Shapes.IndexedTriangleSet(%d).Vertices.v(%d)#p", run, v-1)));
								if x > maxX then
									maxX = x;
								end;
								if x < minX then
									minX = x;
								end;
							end
							--print("---> "..maxX.."/"..minX);
							maxDim = maxX - minX;
							centerOffset = maxX - maxDim/2;
							--maxDim = 2*math.min(math.abs(maxX), math.abs(minX));
							break;
						end
					end
					run = run + 1;
				end
			end
		end
	
		--# rest 
		

		local px,py,pz = getWorldTranslation(w.driveNode);		
		local str = getXMLString(xmlFile, string.format("vehicle.wheels.wheel(%d)#repr", i-1) );
		local str2 = getXMLString(xmlFile, string.format("vehicle.wheels.wheel(%d)#driveNode", i-1) );
		if str2 ~= nil then
			str = str2;
		end
		--print("wheel info from xmlFile("..tostring(xmlFile)..") ".. tostring(str) .. "  for: " .. string.format("vehicle.wheels.wheel(%d)", i-1));

		
		local track = {};
		track.doMeasure = true;
		track.size = maxDim;
		if maxDim == -1 then 
			track.size = 0;
		end;
		track.oldPos = {px, py, pz};
		track.wheelId = i;
		track.iRun = 1;
		track.iMax = 100;
		track.fileIds = {};
		for k = 1,track.iMax do 
			table.insert( track.fileIds, 0 );
		end;
		track.timeOnField = 0;
		track.maxTimeOnField = 10000;
		track.dirtFactorDry = 0.5;
		track.dirtFactorWet = 1.0;
		track.dirtSpeedFactor = 2;
		track.cleanFactorDry = 1.0;
		track.cleanFactorWet = 0.5;
		track.cleanSpeedFactor = 2;
		table.insert(self.wt.tracks, track);
		
		
		if maxDim ~= -1 then 
			-- which comp.? 
			local compNr;
			local begin = string.find(str, ">", nil, true);
			if begin ~= nil then
				compNr = string.byte(str)-48;
			else
				compNr = 0;
			end;
			px,py,pz = worldToLocal( self.components[compNr+1].node, px, py, pz );
			
			px = px + centerOffset;
			
			local dustnode = createTransformGroup( string.format('dustnode%d',i) );
			link( self.components[compNr+1].node, dustnode );
			--setTranslation( dustnode, px, py-w.radius*0.95, pz );
			setTranslation( dustnode, px, py-w.radius+0.05, pz );
			setRotation( dustnode, 0, 0, 0 );		
			
			local iRun = 0;
			local childId = getChildAt( self.components[compNr+1].node, iRun );
			while childId ~= nil do
				childId = getChildAt( self.components[compNr+1].node, iRun );
				if childId == dustnode then
					break;
				end;
				iRun = iRun + 1;
			end
			local reprStr = string.format("%d>%d", compNr, iRun);
			
			local node2 = createTransformGroup( string.format('tracknode',i) );
			link( dustnode, node2 );
			setTranslation( node2, 0, -0.05, 0);
			local reprStr2 = string.format("%d>%d|0", compNr, iRun);
		
			table.insert(tmpDustNodeRefString, i, string.format("%d>%d", compNr, iRun));
			table.insert(tmpDustnodes, i, dustnode);
			
			xmlFileT:write('        <drivingParticle node="'..tostring(reprStr)..'"'..
											' node2="'..tostring(reprStr2)..'"'..
											' position="'..tostring(0).." "..tostring(0).." "..tostring(0)..'"'..
											' rotation="'..tostring(0).." "..tostring(0).." "..tostring(0).. '"'..
											' file="dustParticleSystem.i3d" />\n');
			
			--# create a workingarea for fruit destruction'
			if self.wt.tracks[i].size > 0 then
				local area = {};
				area.wheelIdx = i;
				area.oldPos = {0, 0, 0};
				area.s = createTransformGroup( string.format('destructS%d',i) );
				area.w = createTransformGroup( string.format('destructW%d',i) );
				area.h = createTransformGroup( string.format('destructH%d',i) );		
				link( dustnode, area.s );
				link( dustnode, area.w );
				link( dustnode, area.h );
				
				--print("destruction size = "..size);
				local size = self.wt.tracks[i].size/2;
				setTranslation( area.s, -size, 0.1, -size );	
				setTranslation( area.w,  size, 0.1, -size );	
				setTranslation( area.h, -size, 0.1,  size );			
				table.insert(self.destructAreas, i, area);
			end;
		end;
		
	end;
	xmlFileT:write("    </drivingParticles>\n    <clumbParticles>\n");
	for i,w in pairs(self.wheels) do
		if self.wt.tracks[i].size > 0 then
			
			local nodeClumb = createTransformGroup( string.format('clumbnode',i) );
			link( tmpDustnodes[i], nodeClumb );
			setTranslation( nodeClumb, 0, self.wt.tracks[i].size/2, -self.wt.tracks[i].size/2);
			setRotation(nodeClumb, Utils.degToRad(0), 0,0);		-- 60
			
			--local wx,wy,wz = getWorldTranslation(nodeClumb);
			--worldToLocal(self.rootNode, wx,wy,wz);
			
			local reprStrClumb = string.format("%s|4", tmpDustNodeRefString[i] );	
			xmlFileT:write('        <clumbParticle '..'node="'..tostring(reprStrClumb)..'"'..
											' position="'..tostring(0).." "..tostring(0).." "..tostring(0)..'"'..
											' rotation="'..tostring(0).." "..tostring(0).." "..tostring(0).. '"'..
											' file="clumbParticleSystem.i3d" />\n');			
			
		end
	end
	xmlFileT:write("    </clumbParticles>\n</tmpConfig>");
	xmlFileT:close();	

	if not hasManualDefinition then
		delete(i3dID);
	end
	
	-- read back
	self.drivingParticles = {};	
	self.clumbParticles = {};	
	local xmlFile2 = loadXMLFile("TempConfig", self.tmpXmlFile);
	local run = 0;
	for i,w in pairs(self.wheels) do	
		if self.wt.tracks[i].size > 0 then
			local key = string.format("tmpConfig.drivingParticles.drivingParticle(%d)", run);
			if not hasXMLProperty(xmlFile2, key) then
				print("Did not found key: " .. tostring(key) );
				break;
			end;
			local particle = {};
			particle.ps = {};
			Utils.loadParticleSystem(xmlFile2, particle.ps, key, self.components, false, nil, self.dsw_directory);		
			particle.node = Utils.indexToObject( self.components, getXMLString(xmlFile2, string.format("tmpConfig.drivingParticles.drivingParticle(%d)#node", run) ) );
			particle.node2 = Utils.indexToObject( self.components, getXMLString(xmlFile2, string.format("tmpConfig.drivingParticles.drivingParticle(%d)#node2", run) ) );
			particle.a = false;
			particle.minSpeed = 2/3600;
			particle.maxSpeed = 40/3600;
			particle.wheelId = i;
			table.insert(self.drivingParticles, i, particle);
			local x,y,z = getWorldTranslation(particle.node);
			self.wt.tracks[i].oldPos = {x,y,z};
			
			--
			local key = string.format("tmpConfig.clumbParticles.clumbParticle(%d)", run);
			if not hasXMLProperty(xmlFile2, key) then
				print("Did not found key: " .. tostring(key) );
				break;
			end;			
			local clumb = {};
			clumb.ps = {};
			--local clumbNode = Utils.indexToObject( self.components, getXMLString(xmlFile2, string.format("tmpConfig.clumbParticles.clumbParticle(%d)#node", i-1) ) );
			Utils.loadParticleSystem(xmlFile2, clumb.ps, key, self.components, false, nil, self.dsw_directory);
			
			refNode = Utils.indexToObject( self.components, string.format("%s|4", tmpDustNodeRefString[i] ) );
			--local wx,wy,wz = localDirectionToWorld(refNode, 0, 0, 1);
			--setDirection(clumb.ps[1].shape, wx,wy,wz, 0,1,0 );

			--setScale(clumb.ps[1].shape, self.wt.tracks[i].size, self.wt.tracks[i].size, 1);
			setScale(refNode, self.wt.tracks[i].size, self.wt.tracks[i].size, self.wt.tracks[i].size);
			
			clumb.a = false;
			clumb.minSpeed = 2/3600;
			clumb.maxSpeed = 40/3600;
			clumb.wheelId = i;			
			table.insert(self.clumbParticles, i, clumb);
			
			run = run + 1;
		end
	end	
	delete(xmlFile2);
end;

function DustWheels:delete()
	if self.hasWheels then
		for _, particle in pairs(self.drivingParticles) do
			Utils.deleteParticleSystem(particle.ps);
		end
		for _, clumb in pairs(self.clumbParticles) do
			Utils.deleteParticleSystem(clumb.ps);
		end
	end
end;

function DustWheels:readStream(streamId, connection)
	if self.hasWheels then
		for i,track in pairs(self.wt.tracks) do	
			if track.size > 0 then
				track.timeOnField = streamReadFloat32(streamId);
			end
		end;
	end
end;

function DustWheels:writeStream(streamId, connection)
	if self.hasWheels then
		for i,track in pairs(self.wt.tracks) do	
			if track.size > 0 then
				streamWriteFloat32(streamId, track.timeOnField);
			end
		end;
	end;
end;

function DustWheels:mouseEvent(posX, posY, isDown, isUp, button)
end;

function DustWheels:keyEvent(unicode, sym, modifier, isDown)
end;

function DustWheels:update(dt)
    if self:getIsActive() then
        if self.isClient then	
			--print("DustWheels:update("..dt..")");
		end
	end
	if not self.hasWheels then
		return;
	end;
	

end;


function DustWheels:updateTick(dt)

	
	if not self.hasWheels then
		return;
	end;
	
	local lastkmh = math.abs( self.lastSpeed*3600 );
	local lastSpeed = math.abs( self.lastSpeed );
	
	local isDry = not(g_currentMission.environment.lastRainScale <= 0.1 and g_currentMission.environment.timeSinceLastRain > 30);
	
	--#
    if self:getIsActive() and self.isServer then
        --if self.isClient then	
			--print("DustWheels:updateTick("..dt..")");
			--local areaStr = '';
			--local trackStr = '';
			for i, track in pairs(self.wt.tracks) do
				--areaStr = areaStr .. "i="..i;
				if track.size > 0 then
				
					local x,y,z = getWorldTranslation(self.drivingParticles[i].node);
					
					-- check if 'active'
					if self.wheels[i].hasGroundContact then
					--if true then
					
						--#
						--local cultivVal = Utils.getDensity(g_currentMission.terrainDetailId, g_currentMission.cultivatorChannel, x, z, x+0.05, z, x, z+0.05);
						--local ploughVal = Utils.getDensity(g_currentMission.terrainDetailId, g_currentMission.ploughChannel, x, z, x+0.05, z, x, z+0.05);
						--local sowingVal = Utils.getDensity(g_currentMission.terrainDetailId, g_currentMission.sowingChannel, x, z, x+0.05, z, x, z+0.05);

						local id = g_currentMission.terrainDetailId;
						local area = self.destructAreas[i];
						local x1,y1,z1 = getWorldTranslation(area.s);
						local x2,y2,z2 = getWorldTranslation(area.w);
						local x3,y3,z3 = getWorldTranslation(area.h);
						local xx, zz, widthX, widthZ, heightX, heightZ = Utils.getXZWidthAndHeight(id, x1, z1, x2, z2, x3, z3);
						local dens1 = getDensityParallelogram(id, xx, zz, widthX, widthZ, heightX, heightZ, g_currentMission.cultivatorChannel, 1);						
						local dens2 = getDensityParallelogram(id, xx, zz, widthX, widthZ, heightX, heightZ, g_currentMission.ploughChannel, 1);	
						local dens3 = getDensityParallelogram(id, xx, zz, widthX, widthZ, heightX, heightZ, g_currentMission.sowingChannel, 1);	
						
						--areaStr = areaStr .. string.format("   %5.2f %5.2f", cultivVal, dens1);
						
						local sum = dens1 + dens2 + dens3;
						if sum > 0 then
							if lastkmh > 2 then
								local fac = 1;
								if isDry then
									fac = track.dirtFactorDry;									
								else
									fac = track.dirtFactorWet;
								end								
								track.timeOnField = track.timeOnField + dt*fac; --*track.dirtSpeedFactor*lastkmh; 
								if track.timeOnField > track.maxTimeOnField then
									track.timeOnField = track.maxTimeOnField;
								end;
							end
						else
							--print("MINUS1");
							if lastkmh > 2 then
								local fac = 1;
								if isDry then
									fac = track.cleanFactorDry;									
								else
									fac = track.cleanFactorWet;
								end
								track.timeOnField = track.timeOnField - dt*fac*track.cleanSpeedFactor*math.min(1,(lastkmh/40));
								if track.timeOnField < 0 then
									track.timeOnField = 0;
								end;
							end
						end		

						--# check for fruits
						local fdens = 0;
						local foundfruits = {};
						for index,fruit in pairs(g_currentMission.fruits) do
							local fruitDesc = FruitUtil.fruitIndexToDesc[index];
							if fruitDesc.needsSeeding then	
								fdens = fdens + Utils.getFruitArea(index, x1, z1, x2, z2, x3, z3);						
								table.insert(foundfruits, index, fruit);
							end
						end;
						 
						--areaStr = areaStr .. string.format("   %5.2f %5.2f %5.2f %5.2f  %s   %5.2f", dens1, dens2, dens3, sum, tostring(track.timeOnField), fdens);

						
						--#
						if track.timeOnField > 0 and sum == 0 and fdens == 0 then 
							local rx,ry,rz = getWorldRotation(self.drivingParticles[i].node);
							local h = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x,y,z) + 0.02;
						
							
							local d = {x-track.oldPos[1], y-track.oldPos[2], z-track.oldPos[3]};
							local dl = Utils.vector3Length(d[1], d[2], d[3]);
							local dn = {d[1]/dl, d[2]/dl, d[3]/dl};
							local newPos = {track.oldPos[1], track.oldPos[2], track.oldPos[3]};
							--print(i.." dl="..dl.." track.size="..track.size);
							
							while dl > track.size do
							--if dl > track.size then
							
								--newPos = {x,y,z};
								local step = track.size / 2;
								newPos = {newPos[1]+step*dn[1], newPos[2]+step*dn[2], newPos[3]+step*dn[3]};
								--newPos = {newPos[1]+dn[1], newPos[2]+dn[2], newPos[3]+dn[3]};
								
								track.oldPos = {newPos[1], newPos[2], newPos[3]};
								dl = dl - track.size;
								--print(dl);
								
								local x2,y2,z2 = getWorldTranslation(self.drivingParticles[i].node2);
								
								local hfin = y2 + 0.05;
								if h > hfin then
									hfin = h;
								end;
								
								--local dx,dy,dz = localDirectionToWorld(self.drivingParticles[i].node, 0,-1,0);
								--local dx2,dy2,dz2 = localDirectionToWorld(self.drivingParticles[i].node, track.size/2,0,0);
								--local down = {dx,dy,dz};
								--local hits = raycastClosest( x+dx2,y+dy2+0.1,z+dz2, down[1],down[2],down[3], "rcCallBack", 10, self );
								--print(i.." hits="..tostring(hits).." "..tostring(self.actualHitDist));
								--if hits > 0 then
								--	if self.actualHitDist > 0 then
								--		hfin = y - (0.1 + down[2]*self.actualHitDist) + 0.05;
								--	end;
								--end
								
								local alpha = math.max(0.2, track.timeOnField / track.maxTimeOnField);
								alphaInt = math.floor(alpha * 255 + 0.5);

								if self.isServer then
									--print("-s>");
									if MapDirtTrack ~= nil then
										if MapDirtTrack.MapDirtTrack ~= nil then
											if MapDirtTrack.MapDirtTrack.createNewDirtTrack ~= nil then
												MapDirtTrack.MapDirtTrack:createNewDirtTrack(self, i, x, hfin, z, alphaInt);
												g_server:broadcastEvent(MapDirtTrack.CreateNewDirtTrackEvent:new(self, i, x, hfin, z, alphaInt), nil, nil, self);
											end;
										end;
									end;
								--else
								--	print("-c>");
								--	g_client:getServerConnection():sendEvent(MapDirtTrack.CreateNewDirtTrackEvent:new(self, i));
								end;								
								--self:createNewDirtTrack(i);
								
								--MapDirtTrack.CreateNewDirtTrackEvent.sendEvent(vehicle, wheelId);
								
								--if track.fileIds[track.iRun] ~= 0 then
								--	delete( track.fileIds[track.iRun] );
								--end;
								--track.fileIds[track.iRun] = dirtID;
								--track.iRun = track.iRun + 1;
								--if track.iRun > track.iMax then
								--	track.iRun = 1;
								--end;
							end;
							--track.oldPos = {newPos[1], newPos[2], newPos[3]};
							
						else
							track.oldPos = {x,y,z};
						end;
						
							
						--# destroy only if not on cult/plough/sow
						--if sum == 0 then
							--local area = self.destructAreas[i];
							--local x,y,z = getWorldTranslation(area.s);
							--local x1,y1,z1 = getWorldTranslation(area.w);
							--local x2,y2,z2 = getWorldTranslation(area.h);
							--if Utils.vector3Length(area.oldPos[1]-x1, area.oldPos[2]-y1, area.oldPos[3]-z1) > track.size then
								--area.oldPos = {x1,y1,z1};
								--for index,fruit in pairs(g_currentMission.fruits) do
								--for index,fruit in pairs(foundfruits) do								
									--local fruitDesc = FruitUtil.fruitIndexToDesc[index];
									--if fruitDesc.needsSeeding then
										--local area = Utils.cutFruitArea(index, x1, z1, x2, z2, x3, z3);
										--Utils.updateFruitCutShortArea(index, x1, z1, x2, z2, x3, z3, 0);
									--end;
								--end;
							--end;	
						--end;		
						
					else
						track.oldPos = {x,y,z};
						--print("MINUS3");
						if lastkmh > 2 then
							local fac = 1;
							if isDry then
								fac = track.cleanFactorDry;									
							else
								fac = track.cleanFactorWet;
							end						
							track.timeOnField = track.timeOnField - dt*fac*track.cleanSpeedFactor*math.min(1,(lastkmh/40));						
							if track.timeOnField < 0 then
								track.timeOnField = 0;
							end;
						end
					end;
				end
				--areaStr = areaStr .. string.format("\n");
			end
			--print(areaStr);
		--end
	end
	
	--#
	if self:getIsActive() then
		if self.isClient then
			for idx,area in pairs(self.destructAreas) do
				--if self:getIsAreaActive(area) then
					local x,y,z = getWorldTranslation(area.s);
					local x1,y1,z1 = getWorldTranslation(area.w);
					local x2,y2,z2 = getWorldTranslation(area.h);
					
					local x,y,z = getWorldTranslation(self.wheels[area.wheelIdx].repr);
					y = y - self.wheels[area.wheelIdx].radius;				
					local height = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x,y,z); 
					local hasGroundContact = math.abs(y - height) < self.drivingParticleThreshold;					
					
					if hasGroundContact then
						if Utils.vector3Length(area.oldPos[1]-x, area.oldPos[2]-y, area.oldPos[3]-z) > self.wt.tracks[idx].size then
							area.oldPos = {x,y,z};
							--local str = '';
							--print("---------------------");
							for index,fruit in pairs(g_currentMission.fruits) do
								local fruitDesc = FruitUtil.fruitIndexToDesc[index];
								--print(tostring(index).." "..tostring(fruit).." "..tostring(fruitDesc.name).." "..tostring(fruitDesc.needsSeeding).." "..tostring(fruitDesc.cutState));
								if fruitDesc.needsSeeding then --and not string.match(fruitDesc.name, 'grass') then
									--str = str .. " " .. fruitDesc.name;
									--if string.match( fruitDesc.name, 'potato' ) or string.match( fruitDesc.name, 'sugarBeet' ) then
										
									--else
										Utils.cutFruitArea(index, x, z, x1, z1, x2, z2);
										Utils.updateFruitCutShortArea(index, x, z, x1, z1, x2, z2, 0);			
										--self:cutFruitArea_forced(index, x, z, x1, z1, x2, z2);
										--Utils.updateFruitCutShortArea(index, x, z, x1, z1, x2, z2, 0);			
									--end;
								end;
							end;
							--print(str);
							
							--Utils.updateFruitPreparerArea(index, x, z, x1, z1, x2, z2, 0, 0, 0, 0, 0, 0);
							--Utils.updateDestroyCommonArea( x, z, x1, z1, x2, z2, true ); 
							--self:updateDestroyCommonAreaWoW( x, z, x1, z1, x2, z2, true ); 
						end;
					end;
				--end;
			end;
		end;
	end;	

		
	--#

	local check = false;
	self.dtTmp = self.dtTmp + dt;
	if self.dtTmp > self.dtDelta then
		check = true;
		self.dtTmp = 0;
	end;
	if self.isClient then
		if self:getIsActive() and check then
		--if check then 
			local isVisible = false;
			local camera = getCamera();
			if camera ~= 0 then
				local x,y,z = getWorldTranslation(camera);
				local x1,y1,z1 = getWorldTranslation(self.rootNode);
				local distance = Utils.vector3Length(x-x1,y-y1,z-z1);
				isVisible = (distance < 200);
			end;
			
			local dens = 0; 
			if isVisible then
				local dens1=0; local dens2=0; local dens3=0; local dens4=0;
				local x,y,z = getWorldTranslation(self.wheels[2].repr);
				local x1,y1,z1 = getWorldTranslation(self.wheels[1].repr);
				local x2,y2,z2;
				if self.wheels[4] ~= nil then
					x2,y2,z2 = getWorldTranslation(self.wheels[4].repr);
				else
					x2,y2,z2 = getWorldTranslation(self.components[1].node);
				end;
				local id = g_currentMission.terrainDetailId;
				local xx, zz, widthX, widthZ, heightX, heightZ = Utils.getXZWidthAndHeight(id, x, z, x1, z1, x2, z2);
				local area = Utils.getFruitArea(FruitUtil.FRUITTYPE_DUST, x, z, x1, z1, x2, z2);
				--local areaW = Utils.getFruitArea(FruitUtil.FRUITTYPE_WHEAT, x, z, x1, z1, x2, z2);
				--print("FruitUtil.FRUITTYPE_DUST="..tostring(FruitUtil.FRUITTYPE_DUST).." area="..tostring(area));
				--print("FruitUtil.FRUITTYPE_WHEAT="..tostring(FruitUtil.FRUITTYPE_WHEAT).." areaW="..tostring(areaW));
				if area > 0 then
					dens4 = area;
				else
					dens1 = getDensityParallelogram(id, xx, zz, widthX, widthZ, heightX, heightZ, g_currentMission.cultivatorChannel, 1);
					dens2 = getDensityParallelogram(id, xx, zz, widthX, widthZ, heightX, heightZ, g_currentMission.sowingChannel, 1);
					dens3 = getDensityParallelogram(id, xx, zz, widthX, widthZ, heightX, heightZ, g_currentMission.ploughChannel, 1);
				end;
				--if g_currentMission.environment.lastRainScale <= 0.1 and g_currentMission.environment.timeSinceLastRain > 30 then
					dens = dens1+dens2+dens3+dens4;
				--end;
			end;
			
			--dens = 10;
			--print("self.wt.lastTrackTime="..tostring(self.wt.lastTrackTime).." self.time="..tostring(self.time));
			
			
			
            for k, particle in pairs(self.drivingParticles) do		
				local isActive = false;
				local isActiveC = false;
				
				local x,y,z = getWorldTranslation(self.wheels[particle.wheelId].repr);--getWheelShapeContactPoint(self.wheels[particle.wheel].node, self.wheels[particle.wheel].wheelShape);
				y = y - self.wheels[particle.wheelId].radius;
				
				local height = getTerrainHeightAtWorldPos(g_currentMission.terrainRootNode, x,y,z); 
				local hasGroundContact = math.abs(y - height) < self.drivingParticleThreshold;				
				
				if true then --  if dens > 0 then
					if math.abs(self.lastSpeedReal) > math.abs(particle.minSpeed) then --and (particle.direction == 0 or (particle.direction > 0) == (self.movingDirection > 0)) then															
						isActive = isVisible and hasGroundContact and (not isDry);
						if dens > 0 then
							isActiveC = isVisible and hasGroundContact;
						end;
					end;
					setShaderParameter(self.clumbParticles[k].ps[1].shape, "alphaScale", 1, 0, 0, 1, false);
				elseif self.wt.lastTrackTime ~= nil then --and hasGroundContact 
					if self.wt.lastTrackTime > (self.time - 400) then		
						local a = math.sqrt(self.wt.lastTrackAlpha); --math.max( 1, self.wt.lastTrackAlpha);
						setShaderParameter(self.clumbParticles[k].ps[1].shape, "alphaScale", a, 0, 0, a, false);
						isActiveC = true;
					end
				end;
				
				--isActive = true;
				
				if isActive == true and particle.a == false then
					Utils.setEmittingState(particle.ps, isActive);
					particle.a = true;					
				elseif isActive == false and particle.a == true then
					Utils.setEmittingState(particle.ps, isActive);
					particle.a = false;
				end;
				if isActiveC == true and self.clumbParticles[k].a == false then
					Utils.setEmittingState(self.clumbParticles[k].ps, isActiveC);
					self.clumbParticles[k].a = true;					
				elseif isActiveC == false and self.clumbParticles[k].a == true then
					Utils.setEmittingState(self.clumbParticles[k].ps, isActiveC);
					self.clumbParticles[k].a = false;
				end;		

				
				
            end;
			
		elseif check == true then
			for idx, particle in pairs(self.drivingParticles) do
				if particle.a == true then
					Utils.setEmittingState(particle.ps, false);
					particle.a = false;
				end;
				if self.clumbParticles[idx].a == true then
					Utils.setEmittingState(self.clumbParticles[idx].ps, false);
					self.clumbParticles[idx].a = false;
				end;				
			end;
        end;
    end;
end

function DustWheels:draw()
	if self.hasWheels then
		if self:getIsActive() then
			if self.isClient then	
				for i,track in pairs(self.wt.tracks) do
					if track.measPnt1 ~= nil then
						drawDebugLine( track.measPnt1[1], track.measPnt1[2], track.measPnt1[3], 1,0,0, track.measPnt2[1],track.measPnt2[2],track.measPnt2[3], 0,1,0 );
					end
				end
			end
		end
	end
end;

function DustWheels:setNewLimit(level, percentage, noEventSend)	

end;

function DustWheels:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
    return BaseMission.VEHICLE_LOAD_OK;
end;

function DustWheels:getSaveAttributesAndNodes(nodeIdent)
	local attributes;
    return attributes, nil;
end;

function DustWheels:rcCallBack(hitId, wx, wy, wz, dist)
	--print("hitId="..tostring(hitId).." "..tostring(wx).." "..tostring(wy).." "..tostring(wz));
	if hitId == g_currentMission.terrainRootNode then
		self.actualHitDist = -1;
	else
		self.actualHitDist = dist;
	end
end;




--#
function DustWheels.cutFruitArea_forced(_func, fruitId, startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ, destroySpray, destroySeedingWidth)
	local ids = g_currentMission.fruits[fruitId];
	if ids == nil or ids.id == 0 then
		return 0;
	end
	local id = ids.id;
	local desc = FruitUtil.fruitIndexToDesc[fruitId];
	--print("---------");
	--for i, j in pairs(desc) do
	--	print(tostring(i).." "..tostring(j));
	--end
	local value = desc.cutState+1;
	local x,z, widthX,widthZ, heightX,heightZ = Utils.getXZWidthAndHeight(id, startWorldX, startWorldZ, widthWorldX, widthWorldZ, heightWorldX, heightWorldZ);
	  
	local detailId = g_currentMission.terrainDetailId;
	local spraySum = 0;
	if destroySpray then
		setDensityMaskParams(detailId, "between", desc.minHarvestingGrowthState+1, desc.maxHarvestingGrowthState+1); -- add 1 since growth state 0 has density value 1
		spraySum = setDensityMaskedParallelogram(detailId, x, z, widthX, widthZ, heightX, heightZ, g_currentMission.sprayChannel, 1, id, 0, g_currentMission.numFruitStateChannels, 0);
		setDensityMaskParams(detailId, "greater", 0);
	end
	if desc.useSeedingWidth and (destroySeedingWidth == nil or destroySeedingWidth) then
		setDensityMaskParams(detailId, "between", desc.minHarvestingGrowthState+1, desc.maxHarvestingGrowthState+1); -- add 1 since growth state 0 has density value 1
		local detailTypeValue = 2^g_currentMission.sowingChannel;
		setDensityMaskedParallelogram(detailId, x, z, widthX, widthZ, heightX, heightZ, g_currentMission.terrainDetailTypeFirstChannel, g_currentMission.terrainDetailTypeNumChannels, id, 0, g_currentMission.numFruitStateChannels, detailTypeValue);
		setDensityMaskParams(detailId, "greater", 0);
	end

	  
	-- if no fruit is there, the value is 0 or 1, thus we need to shift by -1, to get values from 0-4, where 0 is no and 4 is full
	setDensityReturnValueShift(id, -1);
	setDensityCompareParams(id, "between", desc.minHarvestingGrowthState+1, desc.maxHarvestingGrowthState+1); -- add 1 since growth state 0 has density value 1
	local pixelsSum, numPixels = setDensityParallelogram(id, x, z, widthX, widthZ, heightX, heightZ, 0, g_currentMission.numFruitStateChannels, value);
	setDensityCompareParams(id, "greater", -1);
	setDensityReturnValueShift(id, 0);
	
	--
	--setDensityTypeIndexCompareMode(id, TYPE_COMPARE_NOTEQUAL);
	--setDensityCompareParams(id, "greater", 0);
	--setDensityParallelogram(id, x, z, widthX, widthZ, heightX, heightZ, 0, g_currentMission.numFruitStateChannels, 0);
	--setDensityCompareParams(id, "greater", -1);
	--setDensityTypeIndexCompareMode(id, TYPE_COMPARE_EQUAL);


	if desc.allowsPartialGrowthState then
		return pixelsSum / desc.maxHarvestingGrowthState, numPixels, spraySum;
	else
		return numPixels, numPixels, spraySum;
	end
end
